-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.modules = q.modules or {};
q.modules["QoL_ClassicThreat-TBC"] = q.modules["QoL_ClassicThreat-TBC"] or {};
local c = q.modules["QoL_ClassicThreat-TBC"];

local barFrames = {};

local barFrameDefaultBottom, barFrameDefaultLeft = 500, 500;
local barFrameDefaultWidth, barFrameDefaultHeight = 200, 100;

local minWidth, minHeight = 100, 50;
local minLeft, maxLeft, minBottom, maxBottom = 0, 1517, 0, 855;

local barHeight, barMarginX, barMarginY = 12, 4, 2;

local DYNAMIC_UNIT_IDS = { "target" }; -- currently "boss mode"

-- TODO: SHOW RAID MARKERS
-- TODO: TARGET ALLY -> SHOW NPC THREAT

function c:GetBarFrameMinWidth()
    return minWidth;
end

function c:GetBarFrameMinHeight()
    return minHeight;
end

local function ConvertColor(colorTable)
    local colors = {};
    if colorTable.r and colorTable.g and colorTable.b then
        -- named properties
        colors = colorTable;
    else
        -- simple array
        colors.r = colorTable[1];
        colors.g = colorTable[2];
        colors.b = colorTable[3];
        colors.a = colorTable[4];
    end
    return colors.r, colors.g, colors.b, colors.a;
end

local function CalculateWidth(barFrame, threatPct)
    local maxWidth = barFrame:GetWidth() - 2 * barMarginX;
    local width = threatPct * maxWidth / 100;
    if width > maxWidth then
        return maxWidth;
    end
    return width;
end

local function IsDynamicUnitId(unitId)
    for _, dynamicUnitId in ipairs(DYNAMIC_UNIT_IDS) do
        if dynamicUnitId == unitId then
            return true;
        end
    end
end

local function GetNextUnitId(barFrame)
    local members, nextId = { unpack(DYNAMIC_UNIT_IDS), unpack(q:GetCurrentMemberNamesAlphabetically()) };
    local isDynamicId = IsDynamicUnitId(barFrame:GetUnitId());

    for index, unitIdOrName in ipairs(members) do
        if (isDynamicId and unitIdOrName == barFrame:GetUnitId()) or (not isDynamicId and unitIdOrName == barFrame:GetUnitName()) then
            if table.getn(members) > index then
                nextId = members[index + 1];
            end
            break;
        end
    end

    nextId = nextId or members[1];
    if IsDynamicUnitId(nextId) then
        return nextId;
    else
        return q:GetCurrentMembers()[nextId];
    end
end

local function Header_OnClick(barFrame, button)
    if button == "LeftButton" then
        -- cycle team members
        barFrame:SetUnitId(GetNextUnitId(barFrame), true);
    elseif button == "RightButton" then
        -- context menu
        ToggleDropDownMenu(1, nil, barFrame.header.dropdown, barFrame.header, 3, -3);
    end
end

function c:AddBarFrame(id)
    id = id or c:GetUnusedBarFrameId();

    -- frame
    local newFrame = CreateFrame("Frame", "ClassicThreat_" .. id, UIParent, BackdropTemplateMixin and "BackdropTemplate");
    newFrame.id = id;
    barFrames[id] = newFrame;

    newFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", barFrameDefaultLeft, barFrameDefaultBottom);
    newFrame:SetSize(barFrameDefaultWidth, barFrameDefaultHeight);
    newFrame:SetFrameLevel(3);
    newFrame:SetMinResize(minWidth, minHeight);
    newFrame.bars = {};

    newFrame:SetBackdrop({
        bgFile = [[Interface\Buttons\WHITE8x8]],
        edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
        tile = true,
        tileSize = 10,
        edgeSize = 10,
        insets = {left = 4, right = 4, top = 4, bottom = 4}
    });
    newFrame:SetBackdropColor(0, 0, 0, 0.3);

    newFrame:SetClipsChildren(true);
    newFrame:SetClampedToScreen(true);
    newFrame:SetResizable(true);
    newFrame:SetMovable(true);
    newFrame:EnableMouse(true);
    newFrame:RegisterForDrag("LeftButton");

    newFrame:SetScript("OnDragStart", function (self)
        self:StartMoving();
    end);
    newFrame:SetScript("OnDragStop", function (self)
        self:StopMovingOrSizing();
        c:SaveBarFrame(self);
    end);
    newFrame:SetUserPlaced(true);

    newFrame.Move = function (self, targetLeft, targetBottom)
        self:ClearAllPoints();
        self:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", targetLeft, targetBottom);
    end
    newFrame.Close = function (self)
        newFrame:Hide();
        c:SaveBarFrame(newFrame);
    end

    -- header
    newFrame.header = CreateFrame("Frame", newFrame:GetName() .. "_Header", newFrame, BackdropTemplateMixin and "BackdropTemplate");
    newFrame.header:SetPoint("TOPLEFT", newFrame, 2, -2);
    newFrame.header:SetPoint("TOPRIGHT", newFrame, -2, -2);
    newFrame.header:SetHeight(16);

    newFrame.header:SetBackdrop({
        bgFile = [[Interface\Buttons\WHITE8x8]],
        tile = true,
        tileSize = 10
    });
    newFrame.header:SetBackdropColor(1, 1, 1, 0.4);

    newFrame.header:SetScript("OnMouseDown", function (self, button)
        Header_OnClick(newFrame, button);
    end);
    newFrame.header:SetScript('OnEnter', function(self)
        q:TransitionMove(self.text, 30, 0);
    end);
    newFrame.header:SetScript('OnLeave', function(self)
        q:TransitionMove(self.text, 17, 0);
    end);

    newFrame.header.icon = CreateFrame("Frame", newFrame.header:GetName() .. "_Icon", newFrame.header, BackdropTemplateMixin and "BackdropTemplate");
    newFrame.header.icon:SetSize(15, 15);
    newFrame.header.icon:SetPoint("TOPLEFT", 0, 0);

    newFrame.header.icon.backdrop = {
        bgFile = [[Interface\Buttons\WHITE8x8]],
        edgeFile = [[Interface\Buttons\WHITE8x8]],
        edgeSize = 1
    };

    newFrame.header.icon.texture = newFrame.header.icon:CreateTexture(newFrame.header.icon:GetName() .. "_Texture", "HIGH");
    newFrame.header.icon.texture:SetSize(15, 15);
    newFrame.header.icon.texture:SetPoint("TOPLEFT", 0.5, -0.5);
    newFrame.header.icon.texture:SetTexture("Interface\\Icons\\INV_Misc_Head_Human_02");

    newFrame.header.text = newFrame.header:CreateFontString(newFrame.header:GetName() .. "_Text", "ARTWORK", "GameFontNormalSmall");
    newFrame.header.text:SetTextColor(1, 1, 1, 1);
    newFrame.header.text:SetPoint("TOPLEFT", newFrame.header, 17, 0);
    newFrame.header.text:SetPoint("BOTTOMLEFT", newFrame.header, 17, 0);

    newFrame.header.text_hidden = newFrame.header:CreateFontString(newFrame.header:GetName() .. "_Text_Hidden", "ARTWORK", "GameFontNormalSmall");
    newFrame.header.text_hidden:SetTextColor(1, 1, 1, 1);
    newFrame.header.text_hidden:SetPoint("TOPLEFT", self.header, -100, 0);
    newFrame.header.text_hidden:SetPoint("BOTTOMLEFT", self.header, -100, 0);

    -- data
    newFrame.data = CreateFrame("Frame", newFrame:GetName() .. "_Data", newFrame, BackdropTemplateMixin and "BackdropTemplate");
    newFrame.data:SetPoint("TOPLEFT", newFrame, 0, 20);
    newFrame.data:SetPoint("BOTTOMRIGHT", newFrame);

    -- close button
    newFrame.closeButton = CreateFrame("Button", newFrame:GetName() .. "_CloseButton", newFrame, "UIPanelCloseButton");
    newFrame.closeButton:SetSize(20, 20);
    newFrame.closeButton:SetPoint("TOPRIGHT");
    newFrame.closeButton:SetFrameLevel(100);
    newFrame.closeButton:SetScript("OnClick", function (self)
        newFrame:Close();
    end);

    -- resize button
    newFrame.resizeButton = CreateFrame("Button", newFrame:GetName() .. "_ResizeButton", newFrame);
    newFrame.resizeButton:SetSize(19, 15);
    newFrame.resizeButton:SetPoint("BOTTOMRIGHT");
    newFrame.resizeButton:SetScript("OnMouseDown", function (self)
        newFrame:StartSizing();
    end);
    newFrame.resizeButton:SetScript("OnMouseUp", function(self)
        newFrame:StopMovingOrSizing();
        newFrame:Update(c:GetThreatCache());
        c:SaveBarFrame(newFrame);
    end)

    newFrame.GetId = function(self)
        return self.id;
    end

    newFrame.GetUnitId = function(self)
        return self.unitId;
    end

    newFrame.GetUnitName = function(self)
        return self.unitName;
    end

    newFrame.Update = function (self, threatCache)
        -- set all bars as inactive (some will be set back to active in this process)
        for _, bar in ipairs(self.bars) do
            bar:SetPending(true);
        end

        local unitId = self:GetUnitId();
        local activeBars = {};
        local unitGuid = UnitGUID(unitId);

        if threatCache and unitGuid then
            local reaction = UnitReaction("player", unitId);
            if IsDynamicUnitId(unitId) and (not reaction or reaction < 4) then
                for memberGuid, memberThreatInfo in pairs(threatCache) do
                    if memberGuid and memberThreatInfo then
                        local threatInfo = memberThreatInfo[unitGuid];
                        if threatInfo and threatInfo.threat > 0 then
                            -- get bar and set data
                            local bar = self:GetBar(threatInfo.testedUnitGuid);
                            bar:Update(threatInfo, threatInfo.testedUnitName, threatInfo.testedUnitGuid);

                            -- prepare bar sorting
                            activeBars[threatInfo.threat] = activeBars[threatInfo.threat] or {};
                            tinsert(activeBars[threatInfo.threat], bar);
                        end
                    end
                end
            elseif threatCache[unitGuid] then
                for targetGuid, threatInfo in pairs(threatCache[unitGuid]) do
                    if threatInfo and threatInfo.threat > 0 then
                        -- get bar and set data
                        local bar = self:GetBar(targetGuid);
                        bar:Update(threatInfo, threatInfo.targetName);

                        -- prepare bar sorting
                        activeBars[threatInfo.threat] = activeBars[threatInfo.threat] or {};
                        tinsert(activeBars[threatInfo.threat], bar);
                    end
                end
            end
        end

        -- sort and align bars
        local headerHeight = self.header:GetHeight();
        local i, sortedBarKeys = 0, q:GetSortedTableKeys(activeBars, true);
        for _, key in ipairs(sortedBarKeys) do
            for _, bar in ipairs(activeBars[key]) do
                local top = headerHeight + 4 + i * (barHeight + barMarginY);
                bar:Move(top, CalculateWidth(self, bar.threatInfo.threat));
                i = i + 1;
            end
        end

        -- reset/hide inactive bars
        for _, bar in ipairs(self.bars) do
            if bar:IsPending() then
                bar:Reset();
            end
        end
    end

    newFrame.SetUnitId = function(self, unitId, isMouseOver)
        if not unitId or (unitId ~= "target" and not UnitExists(unitId)) then
            unitId = "player";
        end

        -- old name transition
        self.header.text_hidden:Show();
        self.header.text_hidden:ClearAllPoints();
        self.header.text_hidden:SetPoint("TOPLEFT", self.header, 30, 0);
        self.header.text_hidden:SetPoint("BOTTOMLEFT", self.header, 30, 0);
        self.header.text_hidden:SetText(self:GetUnitName());
        q:TransitionMove(self.header.text_hidden, self:GetWidth() + 50, 0, 10, function ()
            self.header.text_hidden:Hide();
        end);

        -- new name transition
        self.header.text:ClearAllPoints();
        self.header.text:SetPoint("TOPLEFT", self.header, -100, 0);
        self.header.text:SetPoint("BOTTOMLEFT", self.header, -100, 0);
        if isMouseOver then
            q:TransitionMove(self.header.text, 30, 0, 10);
        else
            q:TransitionMove(self.header.text, 20, 0, 10);
        end

        -- new data
        self.unitId = unitId;
        self.unitName = UnitName(unitId) or q:GetText("Target");

        -- new visuals
        if unitId == "target" then
            if UnitExists("target") then
                self.header.text:SetText(UnitName("target"));
                local r, g, b = UnitSelectionColor("target");
                self.header:SetBackdropColor(r, g, b, 0.4);
            else
                self.header.text:SetText(q:GetText("Target"));
                self.header:SetBackdropColor(1, 1, 1, 0.4);
            end
        else
            self.header.text:SetText(self.unitName);
            local r, g, b = GetClassColor(select(2, UnitClass(unitId)));
            self.header:SetBackdropColor(r, g, b, 0.4);
        end

        -- save new unitId
        c:SaveBarFrame(self);

        -- update/reset bars
        self:Update();
    end
    newFrame:SetUnitId("player");

    newFrame.GetBar = function (self, guid)
        -- get existing, active by guid
        for _, bar in ipairs(self.bars) do
            if bar.guid == guid then
                return bar;
            end
        end

        -- if failed: get inactive
        for _, bar in ipairs(self.bars) do
            if not bar:IsActive() then
                return bar;
            end
        end

        -- create new bar if needed
        local index = table.getn(self.bars);
        local bar = CreateFrame("Frame", newFrame:GetName() .. "_Bar" .. index, self, BackdropTemplateMixin and "BackdropTemplate");
        bar:SetPoint("TOPLEFT", self, "TOPLEFT", 0, - 2 - self.header:GetHeight());
        bar:SetBackdrop({
            bgFile = [[Interface\Buttons\WHITE8x8]],
            edgeFile = [[Interface\Buttons\WHITE8x8]],
            edgeSize = 0.5
        });

        bar.text = bar:CreateFontString(bar:GetName() .. "_FontString", "ARTWORK", "GameFontNormalSmall");
        bar.text:SetTextColor(1, 1, 1, 1);
        bar.text:SetPoint("TOPLEFT", bar, 4, 0);
        bar.text:SetPoint("BOTTOMLEFT", bar, 4, 0);

        bar.SetThreatInfo = function (self, threatInfo)
            self.threatInfo = threatInfo;
        end

        bar.SetGuid = function (self, guid)
            self.guid = guid;
        end

        bar.SetColor = function (self, colorTable)
            self:SetBackdropColor(ConvertColor(colorTable));
        end

        bar.SetBorderColor = function (self, colorTable)
            self:SetBackdropBorderColor(ConvertColor(colorTable));
        end

        bar.SetText = function (self, text)
            self.text:SetText(text);
        end

        bar.SetActive = function(self, value)
            self.isActive = value;
        end

        bar.IsActive = function(self)
            return self.isActive;
        end

        bar.SetPending = function (self, value)
            self.isPending = value;
        end

        bar.IsPending = function (self)
            return self.isPending;
        end

        bar.Move = function (self, targetTop, targetWidth)
            -- TODO (optional): smooth movement
            self:ClearAllPoints();
            self:SetPoint("TOPLEFT", self:GetParent(), "TOPLEFT", barMarginX, -targetTop);
            self:SetSize(targetWidth, barHeight);
        end

        bar.Update = function (self, threatInfo, text, testedUnitGuid)
            self:SetActive(true);
            self:SetPending(false);
            self:SetText(q:MathRound(threatInfo.threat, 0) .. "% - " .. text);
            self:SetThreatInfo(threatInfo);
            self.text:SetTextColor(1, 1, 1, 1);
            if threatInfo.status > 0 then
                self:SetColor(c:GetThreatStatusColors(threatInfo.status, 0.4));-- + threatInfo.status * 0.1));
                if threatInfo.status == 1 then
                --    bar.text:SetTextColor(0, 0, 0, 1);
                end
            elseif testedUnitGuid and testedUnitGuid == UnitGUID("player") then
                self:SetColor({ r = 1, g = 1, b = 1, a = 0.4 });
            else
                self:SetColor({ r = 0, g = 0, b = 0, a = 0 });
            end
            self:SetBorderColor(c:GetThreatStatusColors(0));
            self:Show();
        end

        bar.Reset = function(self)
            self:SetPending(false);
            self:SetActive(false);
            self:SetText("");
            self:SetSize(0, 0);
            self:SetColor({ r = 1, g = 1, b = 1, a = 0 });
            self:SetBorderColor({ r = 1, g = 1, b = 1, a = 0 });
            self:SetThreatInfo(nil);
            self:SetGuid(nil);
            self:Hide();
        end

        tinsert(self.bars, bar);
        bar:SetGuid(guid);
        return bar;
    end

    -- header context menu
    newFrame.header.dropdown = CreateFrame("Frame", newFrame.header:GetName() .. "_Dropdown", newFrame.header, "UIDropDownMenuTemplate");

    -- TODO: NEW RAID MEMBER -> REFRESH
    UIDropDownMenu_Initialize(newFrame.header.dropdown, function(self, level, menuList)
        if (level or 1) == 1 then
            -- boss mode
            local info = UIDropDownMenu_CreateInfo();
            info.isTitle = true;
            info.notCheckable = true;
            info.text = q:GetText("Dynamic");
            UIDropDownMenu_AddButton(info);

            -- target
            info = UIDropDownMenu_CreateInfo();
            info.func = function(self)
                newFrame:SetUnitId(self.value);
                CloseDropDownMenus();
            end;
            info.text = "|cffffffff" .. q:GetText("Target");
            info.value = "target";
            info.checked = (info.value == newFrame:GetUnitId());
            UIDropDownMenu_AddButton(info);

            -- member mode
            info = UIDropDownMenu_CreateInfo();
            info.isTitle = true;
            info.notCheckable = true;
            info.text = q:GetText("Static");
            UIDropDownMenu_AddButton(info);

            for _, unitName in ipairs(q:GetCurrentMemberNamesAlphabetically()) do
                info = UIDropDownMenu_CreateInfo();
                info.func = function(self)
                    if self.value ~= newFrame:GetUnitId() then
                        newFrame:SetUnitId(self.value);
                    end
                    CloseDropDownMenus();
                end;
                info.text = "|cffffffff" .. unitName;
                info.value = q:GetCurrentMembers()[unitName];
                info.checked = (info.value == newFrame:GetUnitId());
                UIDropDownMenu_AddButton(info);
            end
        end
    end, "MENU");

    -- additional
    q:HandleElvUi(newFrame, function() end);

    -- restore last state from closed frames if available
    c:TryRestoreBarFrame(newFrame);
    newFrame:Show();

    -- save frame and return
    c:SaveBarFrame(newFrame);

    return newFrame;
end

function c:UpdateBarFrames(threatCache)
    for id, barFrame in pairs(barFrames) do
        if barFrame:IsVisible() then
            barFrame:Update(threatCache);
        end
    end
end

function c:UpdateBarFrameTargets()
    for id, barFrame in pairs(barFrames) do
        if barFrame:IsVisible() then
            if IsDynamicUnitId(barFrame:GetUnitId()) then
                barFrame:SetUnitId(barFrame:GetUnitId()); -- triggers update
            else
                local unitId = q:GetCurrentMembers()[barFrame:GetUnitName()] or "player";
                if unitId and unitId ~= barFrame:GetUnitId() then
                    barFrame:SetUnitId(unitId);
                end
            end
        end
    end
end